package sam.machine.engine ;      

import java.util.* ; 
import java.util.concurrent.* ; 
import sam.machine.lang.* ; 
import sam.util.* ; 

/**   
  *  ActiveRewrite is Callable agent for applying 
  *  a sam.machine.lang.Rewrite to a branch using staged inference.
  *  @version 9/10/2021  (in progress ...)
    */
public class ActiveRewrite implements ActiveInference { 

   protected final Rewrite rewrite ;             // the rewrite data
   protected final int index ;                   // 1-based index of this rewrite
   protected InferenceManager manager ;          
   protected int U ;                             // # variables in Lterm of rewrite
   protected final Functor [] bindings ;         // variable bindings for Lterm
   protected Cache cache ;   

   /**
     *  Construct a static ActiveRewrite which knows its
     *  static Rewrite lang component, and its InferenceManager.
     */
   protected ActiveRewrite(Rewrite r, InferenceManager mngr, int indx) { 
      this.rewrite = r ; 
      this.index = indx ;  
      this.manager = mngr ;        
      this.U = r.U ;               // U == 0 means factual 
      this.bindings =  new Functor[U] ; 
      this.cache = new Cache() ; 
   }  

   /** 
     * Clear entire bindings table 
      */
   protected void clear_bindings() { 
      for (int i = 0 ; i < U ; i++)       
         bindings[i] = null ;    // now unbound
   }

   /** 
     * Clear cache 
      */
   protected void clear_cache() { 
      this.cache.clear() ; 
   }


   /** 
     * Call this active static rewrite for 
     * staged saturation inference. 
      */
   public Cache call() {
      // fill in action here
      return this.cache ; 
   } 


} // end ActiveRewrite

/*  ALGORITHM DESCRIPTIONS

SIMPLE EXAMPLE.
     rewrite 
        f(a) = b. 
     modulates branch fact
        p(f(a),a,f(a)) 
     to produce modulants 
        p(b,a,f(a))  &  p(f(a),a,b)
     in one level and then 
        p(b,a,b)  &  p(b,a,b)
     in the next level.  

     In the 2nd level for modulation the
     recurrent modulant might NOT be add to 
     the Cache.

A complete set of modulants in generated over 
k levels if branch fact P has k matching occurrences of
the Lterm.  


Given branch fact F and R=Rewrite(Lterm,Rterm), 
design how it is that AR=ActiveRewrite(R) should 
act on F.

High level description:

AR searches F for all subterm instances of Lterm.
When a subterm instance is located, the binding 
table is used to generate a new subterm via 
substituting the variable bindings into Rterm.
A tentative new fact F' is generate by replacing 
the matched Lterm by the substituted Rterm.  If F'
is fresh, then F' is added to the return cache list of 
facts.  Between attempts to locate subterm instances
of Lterm the variable binding table is cleared.

(This description is concise enough to implement AR.
However, there would be more, regarding how to add the 
cache of new facts to the branch at the end of staged
inference.)
*/
